/*
 * Sistemas de Tempo-Real, 2003/2004
 * Paulo Pedreiras, Oct.2003
 *
 * Demonstracao do uso do Shark
 *
 * System initialization file
 */

/*
 This file contains the functions needed to initialize the system.
 
 These functions register the following levels:
 
 Level 0: INTDRIVE level for interrupt handling
 Level 1: EDF (Earliest Deadline First) level
 Level 2: CBS (Constant Bandwidth Server) level
 Level 3: RR (Round Robin) level
 Level 4: Dummy level
 
 It can accept these task models:
 
 HARD_TASK_MODEL (wcet+mit) at level 1
 - Provided by EDF module. Periodic and sporadic tasks.
   May execute admission control and wcet check.
 
 SOFT_TASK_MODEL (met, period) at level 2
 - Provided by CBS module. Periodic and sporadic tasks.
   Inserts task activations in the EDF module.
   May execute admission control.
    
 NRT_TASK_MODEL  at level 3

 !!!!NOTE: level 0 is reserved to interrupt handling, and so
           no user tasks should register at this level

 TICK is set to 0 (one-shot timer is used)
*/

/* Select SHaRK 1.5 (comment) or 1.51 (uncomment)*/
#define SHARK151

#include <kernel/kern.h>
#include <kernel/mem.h>
#include <ll/i386/cons.h>

#include <edf/edf/edf.h>
#include <cbs/cbs/cbs.h>
#include <rr/rr/rr.h>
#include <dummy/dummy/dummy.h>
#include <intdrive/intdrive/intdrive.h>

#include <sem/sem/sem.h>
#include <hartport/hartport/hartport.h>
#include <cabs/cabs/cabs.h>

#include <drivers/shark_linuxc26.h>
#include <drivers/shark_pci26.h>
#include <drivers/shark_input26.h>
#include <drivers/shark_keyb26.h>

PID shutdown_task_PID = -1;

   /*+ sysyem tick in us +*/
#define TICK 0 

/*+ RR tick in us +*/
#define RRTICK 10000

/*+ Interrupt Server +*/
#ifdef SHARK151
#define INTDRIVE_Q 1000
#define INTDRIVE_U 0.1*MAX_BANDWIDTH
#define INTDRIVE_FLAG 0
#else
#define INTDRIVE_Q 1000
#define INTDRIVE_T 10000
#define INTDRIVE_FLAG 0
#endif

/* Function prototypes */
void call_shutdown_task(void *arg);
int device_drivers_init();
int device_drivers_close();
void set_shutdown_task();
TASK shutdown_task_body(void *arg);

TIME __kernel_register_levels__(void *arg)
{
  struct multiboot_info *mb = (struct multiboot_info *)arg;

  /**********************/
  /* Scheduling modules */
  /**********************/

#ifdef SHARK151
  INTDRIVE_register_level(INTDRIVE_Q,INTDRIVE_Q,INTDRIVE_U,INTDRIVE_FLAG);
#else
  INTDRIVE_register_level(INTDRIVE_Q,INTDRIVE_T,INTDRIVE_FLAG);
#endif

  /* EDF module. Guarantee and wcet check disabled */
  EDF_register_level(EDF_DISABLE_ALL);

  /* CBS module. Guarantee enabled. Client of level 0 module (EDF) */
  CBS_register_level(CBS_ENABLE_ALL, 1);
  
  /* Round Robin Module. main() scheduled in this module */
  RR_register_level(RRTICK, RR_MAIN_YES, mb);
  
  /* Dummy task. Ensures that there is always a task ready to run */
  dummy_register_level();

  /********************/
  /* Resource modules */
  /********************/

  /* Semaphor module registration */
  SEM_register_module();

  /* Cyclical Asynchronous buffers module registration */
  /* Non-blocking multi-access shared buffer */
  CABS_register_module();

  return TICK;
}

TASK __init__(void *arg)
{
  struct multiboot_info *mb = (struct multiboot_info *)arg;


  HARTPORT_init();

  /* Create the shutdown task. It will be activated at RUNLEVEL SHUTDOWN */
  set_shutdown_task();
  
  /* Init the drivers */
  device_drivers_init();

  /* Set the shutdown task activation */
  sys_atrunlevel(call_shutdown_task, NULL, RUNLEVEL_SHUTDOWN);

  /* Execute application */
  __call_main__(mb);

  return (void *)0;
}

void set_shutdown_task()
{
  /* WARNING: the shutdown task is a background thread. It cannot execute if the system is overloaded */
  NRT_TASK_MODEL nrt;
  
  nrt_task_default_model(nrt);
  nrt_task_def_system(nrt);
  
  shutdown_task_PID = task_create("Shutdown Task",shutdown_task_body,&nrt,NULL);
  if (shutdown_task_PID == NIL) {
    sys_shutdown_message("Error: Cannot create shutdown task\n");
    exit(1);
  }
}

int device_drivers_init()
{

  KEYB_PARMS kparms = BASE_KEYB;

#ifdef SHARK151
  LINUXC26_register_module(TRUE);
#else
  LINUXC26_register_module();
#endif
  
  PCI26_init();
  
  INPUT26_init();
  
  /******************/
  /* Keboard set-up */
  /******************/
  //keyb_def_ctrlC(kparms, NULL);
  KEYB26_init(&kparms);
  
  return 0;
}

int device_drivers_close() {
  
  KEYB26_close();
  INPUT26_close();
  
  return 0;                                                                                                                     
}

#define SHUTDOWN_TIMEOUT_SEC 3

void call_shutdown_task(void *arg) 
{
  struct timespec t;
  
  sys_gettime(&t);
  t.tv_sec += SHUTDOWN_TIMEOUT_SEC;
  
  /* Emergency timeout to exit from RUNLEVEL_SHUTDOWN */
  kern_event_post(&t,(void *)((void *)sys_abort_shutdown),(void *)0);
  
  task_activate(shutdown_task_PID);
}

TASK shutdown_task_body(void *arg) 
{
  device_drivers_close();
  
  sys_shutdown_message("-- S.Ha.R.K. Closed --\n");
  
  sys_abort_shutdown(0);
  
  return NULL;
}

